@@ -2,12 +2,12 @@ from datetime import datetime
 from functools import partial
 from uuid import UUID
 
+from google.protobuf.descriptor import FieldDescriptor
 from google.protobuf.timestamp_pb2 import Timestamp
 
 
 def _py_result(model, r, del_uuid):
-    """Convert the result of a model operation to a python value,
-       and update the object cache in the model."""
+    """Convert the result of a model operation to a python value."""
 
     def pb2py(pb, cls):
         """convert a protobuf to a python object."""
@@ -23,33 +23,6 @@ def _py_result(model, r, del_uuid):
             if isinstance(v, Timestamp):
                 v = v.ToNanoseconds() / 1000000000
             fields[f] = v
-        uuid = fields.get('uuid')
-        if uuid and model.cache_enabled:
-            # handle this as a managed object
-            cur = model._objs.get(uuid)
-            if cur:
-                curpb, curpy = cur
-                for f, v in fields.items():
-                    try:
-                        pbv = getattr(curpb, f, None) if curpb.HasField(f) else None
-                    except ValueError:
-                        pbv = getattr(curpb, f, None)
-                    if isinstance(pbv, Timestamp):
-                        pbv = pbv.ToNanoseconds() / 1000000000
-                    if v != pbv:
-                        pyv = getattr(curpy, f, None)
-                        if pyv and pbv != pyv:
-                            # This field is dirty.  We may want different
-                            # behavior here someday?  For now overwrite it.
-                            setattr(curpy, f, v)
-                        else:
-                            setattr(curpy, f, v)
-                curpb.CopyFrom(pb)
-            else:
-                curpy = cls(**fields)
-                model._objs[uuid] = (pb, curpy)
-            return curpy
-        # no uuid, just return the object unmanagd.
         return cls(**fields)
 
     if isinstance(r, list):
@@ -62,9 +35,7 @@ def _py_result(model, r, del_uuid):
         if cls:
             return pb2py(r, cls)
         if del_uuid:
-            # this was a delete operation; delete our cached object data
-            cur = model._objs.pop(del_uuid, None)
-            return cur[1] if cur else None
+            return None
         if len(r.DESCRIPTOR.fields) == 0:
             # This is just an empty response, we simply return None.
             return
@@ -104,7 +75,11 @@ def _pb_request(reqtype, **kwargs):
                             if nt.message_type:
                                 getattr(lv, nk).MergeFrom(nv)
                             elif nv is not None:
-                                setattr(lv, nk, nv)
+                                field = x.DESCRIPTOR.fields_by_name.get(nk)
+                                if field and field.label == FieldDescriptor.LABEL_REPEATED:
+                                    getattr(lv, nk).MergeFrom(nv)
+                                else:
+                                    setattr(lv, nk, nv)
                 else:
                     # A plain nested type
                     try:
@@ -113,18 +88,27 @@ def _pb_request(reqtype, **kwargs):
                         getattr(req, k).MergeFrom(v)
 
         else:
-            setattr(req, k, v)
+            if isinstance(v, list):
+                getattr(req, k).MergeFrom(v)
+            else:
+                setattr(req, k, v)
     return req
 
 
-async def op(model, op, reqtype, delete=False, _pythonize=True, **kwargs):
+async def op(model, op, reqtype, delete=False, _pythonize=True, handler=None, **kwargs):
     """Wrapper function to perform a GRPC operation on the model. (asyncio)"""
     req = _pb_request(reqtype, **kwargs)
     company_uuid = kwargs.get('company_uuid') or (
         kwargs.get('uuid') if reqtype.__module__ == 'company_pb2' else None)
     md = (('company-uuid', str(UUID(bytes=company_uuid))), ) if company_uuid else None
-    r = await model._client._grpc_op(op, req, collect_results=True, metadata=md)
+    r = await model._client._grpc_op(op,
+                                     req,
+                                     collect_results=True,
+                                     metadata=md,
+                                     handler=handler)
     del_uuid = kwargs.get('uuid') if delete else None
+    if handler:
+        return r
     return _py_result(model, r, del_uuid) if _pythonize else r
 
 
