@@ -137,7 +137,11 @@ class FakePreseemModel:
                       management_ip=kwargs.get('management_ip'),
                       site_uuid=kwargs.get('site_uuid'),
                       status=kwargs.get('status') or 0,
-                      poller_hash=kwargs.get('poller_hash'))
+                      poller_hash=kwargs.get('poller_hash'),
+                      intent_role_uuid=kwargs.get('intent_role_uuid'),
+                      acting_role_uuid=kwargs.get('acting_role_uuid'),
+                      source=kwargs.get('source'),
+                      source_id=kwargs.get('source_id'))
             elems = FakePreseemModel._elements.get(company.uuid)
             if elems is None:
                 elems = FakePreseemModel._elements[company_uuid] = {}
@@ -152,6 +156,7 @@ class FakePreseemModel:
             return [
                 copy.deepcopy(x)
                 for x in (FakePreseemModel._elements.get(company_uuid) or {}).values()
+                if all([getattr(x, k) == v for k, v in kwargs.items()])
             ]
 
         @classmethod
