@@ -134,12 +134,13 @@ class Subscription(object):
 class PreseemGrpcClient(object):
     """Implement the GRPC client to the Preseem API, adapted to asyncio."""
 
-    def __init__(self, api_key, host=None, port=None, compression=False):
+    def __init__(self, api_key, host=None, port=None, compression=False, insecure=None):
         self.api_key = api_key
         self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=10)
         self.host = host or 'nmapi.preseem.com'
         self.port = port or 443
         self.compression = compression
+        self.insecure = insecure
         self._channel = None
         self._metrics_svc = None
         self._nm_svc = None
@@ -191,11 +192,18 @@ class PreseemGrpcClient(object):
             except Exception as err:
                 logging.exception('Error calling grpc service')
 
-    def _grpc_op_collect(self, op, req, metadata):
-        """Run a GRPC operation and collect the results in this thread."""
+    def _grpc_op_collect(self, op, req, metadata, handler=None):
+        """
+        Run a GRPC operation and collect the results in this thread.
+        If a handler function is passed in, that function will get an interator
+        and can process it however it wants; whatever that function returns will be
+        returned to the caller.
+        """
         r = op(req, metadata=metadata)
         try:
             iterator = iter(r)  # this is to detect whether its a sequence
+            if handler:
+                return handler(iterator)
             return [x for x in iterator]  # this does network i/o
         except TypeError:  # not an iterable result, ok
             pass
@@ -207,7 +215,8 @@ class PreseemGrpcClient(object):
                        num_retries=5,
                        timeout=None,
                        collect_results=None,
-                       metadata=None):
+                       metadata=None,
+                       handler=None):
         loop = asyncio.get_event_loop()
         for retry_num in range(num_retries + 1):
             if retry_num > 0:
@@ -219,7 +228,7 @@ class PreseemGrpcClient(object):
                         if collect_results:
                             return await asyncio.wait_for(loop.run_in_executor(
                                 self._executor, self._grpc_op_collect, op, req,
-                                metadata),
+                                metadata, handler),
                                                           timeout=timeout)
                         else:
                             return await asyncio.wait_for(loop.run_in_executor(
@@ -233,7 +242,7 @@ class PreseemGrpcClient(object):
                     if collect_results:
                         x = await loop.run_in_executor(self._executor,
                                                        self._grpc_op_collect, op, req,
-                                                       metadata)
+                                                       metadata, handler)
                     else:
                         x = await loop.run_in_executor(
                             self._executor, functools.partial(op,
@@ -271,6 +280,55 @@ class PreseemGrpcClient(object):
             def __call__(self, context, callback):
                 callback((('api_key', self.api_key), ), None)
 
+        class AuthInterceptor(
+                grpc.UnaryUnaryClientInterceptor,
+                grpc.UnaryStreamClientInterceptor,
+                grpc.StreamUnaryClientInterceptor,
+                grpc.StreamStreamClientInterceptor,
+        ):
+            """Universal interceptor that adds metadata to all RPC types."""
+
+            def __init__(self, plugin):
+                self.plugin = plugin
+
+            def _augment_metadata(self, client_call_details):
+                """Helper: calls your plugin and merges metadata."""
+                md = []
+
+                def callback(metadata, error):
+                    if metadata:
+                        md.extend(metadata)
+
+                # Run the plugin to populate md
+                self.plugin(None, callback)
+
+                # Merge with any existing metadata
+                combined = list(client_call_details.metadata or []) + md
+                return client_call_details._replace(metadata=tuple(combined))
+
+            # Unary → Unary
+            def intercept_unary_unary(self, continuation, client_call_details, request):
+                new_details = self._augment_metadata(client_call_details)
+                return continuation(new_details, request)
+
+            # Unary → Stream
+            def intercept_unary_stream(self, continuation, client_call_details,
+                                       request):
+                new_details = self._augment_metadata(client_call_details)
+                return continuation(new_details, request)
+
+            # Stream → Unary
+            def intercept_stream_unary(self, continuation, client_call_details,
+                                       request_iterator):
+                new_details = self._augment_metadata(client_call_details)
+                return continuation(new_details, request_iterator)
+
+            # Stream → Stream
+            def intercept_stream_stream(self, continuation, client_call_details,
+                                        request_iterator):
+                new_details = self._augment_metadata(client_call_details)
+                return continuation(new_details, request_iterator)
+
         dst = '{}:{}'.format(self.host, self.port)
         while self._nm_svc is None:
             try:
@@ -281,12 +339,13 @@ class PreseemGrpcClient(object):
                      15000),  # keepalive ping timeout after 15s
                     ('grpc.keepalive_permit_without_calls',
                      True),  # send keepalives when no grpc calls
-                    ('grpc.http2.max_pings_without_data', 0)
+                    ('grpc.http2.max_pings_without_data', 0),
+                    ('grpc.max_receive_message_length', 10 * 1024 * 1024),
                 ]  # allow unlimited pings without data
                 if self.compression:
                     grpc_opts.append(
                         ('grpc.default_compression_algorithm', grpc.Compression.Gzip))
-                if self.api_key:
+                if self.api_key and not self.insecure:
                     call_creds = grpc.metadata_call_credentials(
                         ApiKeyCallCredentials(self.api_key))
                     try:
@@ -300,6 +359,10 @@ class PreseemGrpcClient(object):
                     channel = grpc.secure_channel(dst, channel_creds, options=grpc_opts)
                 else:
                     channel = grpc.insecure_channel(dst, options=grpc_opts)
+                    if self.api_key:
+                        channel = grpc.intercept_channel(
+                            channel,
+                            AuthInterceptor(ApiKeyCallCredentials(self.api_key)))
                 self._channel = channel
                 self._metrics_svc = metrics_pb2_grpc.NetworkMetricsStub(channel)
                 self._nm_svc = network_metadata_pb2_grpc.NetworkMetadataServiceStub(
