@@ -9,15 +9,30 @@ from grpc import RpcError
 
 from preseem import NetworkMetadataReference
 from preseem.source_model import AccessPoint, Olt, Router, Service
-from preseem_protobuf.model.service_pb2 import SetAttachmentReq
+from preseem_protobuf.model.common_pb2 import ElementStatus
+from preseem_protobuf.model.service_pb2 import Attachment, SetAttachmentReq
 
 # STM-6039 Only allow the sync to run once at a time, to prevent duplicate
 # entities from being created.
 sync_lock = None
 
+
 # This object is used to pass element items to the sync_preseem_model
 # algorithm for a given source and intent role.
-ElementSyncItem = namedtuple('ElementSyncItem', ('id', 'name', 'site', 'host'))
+class ElementSyncItem:
+
+    def __init__(self, id, name, site, host):
+        self.id = id
+        self.name = name
+        self.site = site
+        self.host = host
+        self._site_uuid = None
+
+    def __repr__(self):
+        clsname = self.__class__.__name__
+        attrs = ', '.join(f"{k}={v!r}" for k, v in self.__dict__.items() if k[0] != '_')
+        return f"{clsname}({attrs})"
+
 
 _account_field_map = {
     'type': lambda x: x.type,
@@ -33,7 +48,7 @@ _package_field_map = {
     'upload_rate': lambda x: x.upload_rate or None,
 }
 
-_elem_field_map = {
+_elem_disc_field_map = {
     'name': lambda x: x.name,
     'management_ip': lambda x: x.host,
     'site_uuid': lambda x: x._site_uuid,
@@ -50,6 +65,14 @@ _service_field_map = {
     'source_url': lambda x: x._url,
 }
 
+_elem_field_map = {
+    'name': lambda x: x.name,
+    'management_ip': lambda x: x.host,
+    'site_uuid': lambda x: x._site_uuid,
+    'intent_role_uuid': lambda x: x.role_uuid.bytes,
+    'poller_hash': lambda x: x.poller_hash,
+}
+
 
 def has_any_fields_set(message):
     for field_desc, value in message.ListFields():
@@ -75,8 +98,31 @@ async def call_and_log_error(co):
     return None
 
 
-async def _sync_objects(model, company_uuid, source, bss_objs, model_objs, createrpc,
-                        updaterpc, field_map, deactivate, noact):
+async def run_limited(coros, limit: int):
+    """Run a list of coroutines with a maximum concurrency limit."""
+    sem = asyncio.Semaphore(limit)
+
+    async def guarded(coro):
+        async with sem:
+            return await coro
+
+    # schedule everything, but only `limit` run at a time
+    return await asyncio.gather(*(guarded(c) for c in coros), )
+
+
+async def _sync_objects(model,
+                        company_uuid,
+                        source,
+                        bss_objs,
+                        model_objs,
+                        createrpc,
+                        updaterpc,
+                        field_map,
+                        deactivate,
+                        noact,
+                        cfg=None,
+                        init_fields=None,
+                        concurrency=10):
     """
     Function to sync python bss-sourced objects to model objects.
     Pass in the model and the company_uuid and source we are working on.
@@ -88,6 +134,8 @@ async def _sync_objects(model, company_uuid, source, bss_objs, model_objs, creat
     """
     new_ids = set(bss_objs)
     new_objs = []
+    modcos = []
+    delcos = []
     for mobj in model_objs:
         if mobj.source != source:
             continue  # ignore objects owned by other sources
@@ -105,7 +153,11 @@ async def _sync_objects(model, company_uuid, source, bss_objs, model_objs, creat
                 if mval != bval:
                     if f == 'attachment':
                         # find what attachment fields changed for logging purposes
+                        batt = Attachment()
+                        batt.CopyFrom(bval)  # make a copy so I don't modify the input
+                        bval = batt
                         achgs = []
+                        attachment_cfg = cfg.get('attachment') if cfg else {}
                         if mval and bval:
                             for sf in bval.DESCRIPTOR.fields:
                                 smval = getattr(mval, sf.name)
@@ -116,6 +168,13 @@ async def _sync_objects(model, company_uuid, source, bss_objs, model_objs, creat
                                 if sf.name == 'network_prefixes':
                                     smval = set([ip_network(x) for x in smval])
                                     sbval = set([ip_network(x) for x in sbval])
+                                field_cfg = attachment_cfg.get(sf.name)
+                                if field_cfg and field_cfg.get('ignore'):
+                                    # To leave the field as it was, we have to set it
+                                    # to what is in model-server now,
+                                    # because the attachment is set as one full object.
+                                    setattr(bval, sf.name, smval)
+                                    sbval = smval
                                 if smval != sbval:
                                     achgs.append(f"{sf.name}[{smval}->{sbval}]")
                         elif bval and not mval:
@@ -146,38 +205,107 @@ async def _sync_objects(model, company_uuid, source, bss_objs, model_objs, creat
                         if attachment:
                             set_att_req.attachment.CopyFrom(attachment)
                         mods['attachment_req'] = set_att_req
-                    await call_and_log_error(
-                        updaterpc(company_uuid=company_uuid, uuid=mobj.uuid, **mods))
+                    modcos.append(
+                        call_and_log_error(
+                            updaterpc(company_uuid=company_uuid, uuid=mobj.uuid,
+                                      **mods)))
         elif deactivate:  # sync deletes unless explicitly told not to
             if not mobj.inactive:
                 logging.info("sync_preseem_model(%s): deactivate %s %s", source,
                              type(mobj).__name__.lower(), UUID(bytes=mobj.uuid))
                 if not noact:
-                    await call_and_log_error(
-                        updaterpc(company_uuid=company_uuid,
-                                  uuid=mobj.uuid,
-                                  inactive=True))
+                    delcos.append(
+                        call_and_log_error(
+                            updaterpc(company_uuid=company_uuid,
+                                      uuid=mobj.uuid,
+                                      inactive=True)))
+
+    # call modify and delete ops concurrently
+    if modcos:
+        logging.info(f"Modify {len(modcos)} objects")
+        await run_limited(modcos, concurrency)
+        logging.info(f"Modified {len(modcos)} objects")
+    if delcos:
+        logging.info(f"Deactivate {len(delcos)} objects")
+        await run_limited(delcos, concurrency)
+        logging.info(f"Deactivated {len(delcos)} objects")
+
+    # Need to do these concurrent too, easy enough, will need a new helper fn.
+    addcos = []
     for obj_id in new_ids:
         bobj = bss_objs.get(obj_id)
         if bobj:
             logging.info('sync_preseem_model(%s): create %s', source, bobj)
             if not noact:
-                mobj = await call_and_log_error(
-                    createrpc(company_uuid=company_uuid,
-                              source=source,
-                              source_id=bobj.id,
-                              **{
-                                  f: m(bobj)
-                                  for f, m in field_map.items()
-                              }))
-                if mobj:
-                    uuid = UUID(bytes=mobj.uuid)
-                    logging.info('sync_preseem_model(%s): created %s "%s"', source,
-                                 type(mobj).__name__, uuid)
-                    new_objs.append(mobj)
+                addcos.append(
+                    call_and_log_error(
+                        createrpc(company_uuid=company_uuid,
+                                  source=source,
+                                  source_id=bobj.id,
+                                  **{
+                                      f: m(bobj)
+                                      for f, m in field_map.items()
+                                  },
+                                  **{
+                                      f: v
+                                      for f, v in (init_fields or {}).items()
+                                  })))
+
+    if addcos:
+        logging.info(f"Create {len(addcos)} objects")
+        mobjs = await run_limited(addcos, concurrency)
+        for mobj in mobjs:
+            if mobj:
+                uuid = UUID(bytes=mobj.uuid)
+                logging.info('sync_preseem_model(%s): created %s "%s"', source,
+                             type(mobj).__name__, uuid)
+                new_objs.append(mobj)
+        logging.info(f"Created {len(addcos)} objects")
     return new_objs
 
 
+async def _sync_sites(model, company_uuid, source, objs, noact=False):
+    """Make sure model has all sites needed by source objs, and set the site_uuid on each obj."""
+    sites = await model.Site.List(company_uuid=company_uuid)
+    sites = {x.uuid: x for x in sites}
+
+    # Create a map of name -> site.  Handle multiple sites with the same
+    # name by preferring active sites, then taking the first sorted uuid.
+    site_names = {}
+    for site in sorted(sites.values(), key=lambda x: x.uuid):
+        cur_site = site_names.get(site.name)
+        if not cur_site or (cur_site.inactive and not site.inactive):
+            site_names[site.name] = site
+
+    new_sites = set()
+    for sobj in objs:
+        if sobj.site:
+            site = site_names.get(sobj.site)
+            if not site:
+                new_sites.add(sobj.site)
+            elif site.inactive:
+                logging.info('sync_sites(%s): reactivate site "%s"', source, sobj.site)
+                if not noact:
+                    await call_and_log_error(
+                        model.Site.Update(company_uuid=company_uuid,
+                                          uuid=site.uuid,
+                                          inactive=False))
+    for site_name in new_sites:
+        logging.info('sync_sites(%s): create site "%s"', source, site_name)
+        if not noact:
+            site = site_names[site_name] = await call_and_log_error(
+                model.Site.Create(company_uuid=company_uuid, name=site_name))
+            if site:
+                sites[site.uuid] = site_names[site_name] = site
+
+    for sobj in objs:
+        if sobj.site:
+            site = site_names.get(sobj.site)
+            sobj._site_uuid = site.uuid
+        else:
+            sobj._site_uuid = None
+
+
 async def sync_preseem_model(preseem_model,
                              company_uuid,
                              source,
@@ -192,6 +320,7 @@ async def sync_preseem_model(preseem_model,
                              nm_model=None,
                              deactivate=None,
                              noact=False,
+                             cfg=None,
                              verbose=0):
     """
     Update the Preseem model to keep its ElementDiscovery objects in sync with
@@ -214,15 +343,23 @@ async def sync_preseem_model(preseem_model,
         raise ValueError('source cannot be the empty string')
 
     async with sync_lock:
-        model_accounts, model_discs, model_services, model_sites, model_packages = await asyncio.gather(
+        # Collect all discovery objects so we can sync the sites.
+        ne_objs = [
+            obj for d in (aps, routers, olts) if d is not None for obj in d.values()
+        ]
+        for item in items or []:
+            if item.site is not None and (not isinstance(item.site, str)
+                                          or item.site == ''):
+                raise ValueError('invalid string value for site')
+            ne_objs.append(item)
+
+        model_accounts, model_discs, model_services, model_packages, _ = await asyncio.gather(
             preseem_model.Account.List(company_uuid=company_uuid),
             preseem_model.Element.ListElementDiscovery(company_uuid=company_uuid),
             preseem_model.Service.List(company_uuid=company_uuid,
                                        include_attachment=True),
-            preseem_model.Site.List(company_uuid=company_uuid),
-            preseem_model.Package.List(company_uuid=company_uuid))
-
-        model_sites = {x.uuid: x for x in model_sites}
+            preseem_model.Package.List(company_uuid=company_uuid),
+            _sync_sites(preseem_model, company_uuid, source, ne_objs, noact))
 
         role_discs = {}
         for disc in model_discs:
@@ -235,14 +372,6 @@ async def sync_preseem_model(preseem_model,
             model_discs = {x.uuid: x for x in role_discs.get(role.bytes) or []}
             disc_ids = {(x.source, x.source_id): x for x in model_discs.values()}
 
-        # Create a map of name -> site.  Handle multiple sites with the same
-        # name by preferring active sites, then taking the first sorted uuid.
-        site_names = {}
-        for site in sorted(model_sites.values(), key=lambda x: x.uuid):
-            cur_site = site_names.get(site.name)
-            if not cur_site or (cur_site.inactive and not site.inactive):
-                site_names[site.name] = site
-
         # Make sure all sites we reference exist and do some pre-validation.
         new_sites = set()
         item_ids = set()
@@ -252,39 +381,15 @@ async def sync_preseem_model(preseem_model,
                 raise ValueError('invalid string value for source_id')
             if not isinstance(item.name, str) or item.name == '':
                 raise ValueError('invalid string value for name')
-            if item.site is not None and (not isinstance(item.site, str)
-                                          or item.site == ''):
-                raise ValueError('invalid string value for site')
             if item.id in item_ids:
                 dup_ids.add(item.id)
             else:
                 item_ids.add(item.id)
-            if item.site:
-                site = site_names.get(item.site)
-                if not site:
-                    new_sites.add(item.site)
-                elif site.inactive:
-                    logging.info('sync_preseem_model(%s): reactivate site "%s"', source,
-                                 item.site)
-                    if not noact:
-                        await call_and_log_error(
-                            preseem_model.Site.Update(company_uuid=company_uuid,
-                                                      uuid=site.uuid,
-                                                      inactive=False))
         if dup_ids:
             logging.warning('sync_preseem_model(%s): duplicate IDs found: %s', source,
                             dup_ids)
             raise ValueError('duplicate IDs found')
 
-        for site_name in new_sites:
-            logging.info('sync_preseem_model(%s): create site "%s"', source, site_name)
-            if not noact:
-                site = site_names[site_name] = await call_and_log_error(
-                    preseem_model.Site.Create(company_uuid=company_uuid,
-                                              name=site_name))
-                if site:
-                    model_sites[site.uuid] = site_names[site_name] = site
-
         uuid_map = {}
         if role:  # old algorithm
             unref_discs = set(
@@ -296,7 +401,6 @@ async def sync_preseem_model(preseem_model,
                 if disc:
                     uuid_map[item.id] = UUID(bytes=disc.uuid)
                     unref_discs.remove(disc.uuid)
-                    site = model_sites.get(disc.site_uuid) if disc.site_uuid else None
                     changes = []
                     mods = {}
                     if disc.inactive:
@@ -309,9 +413,9 @@ async def sync_preseem_model(preseem_model,
                     if disc.management_ip != host:
                         changes.append(f'management_ip: {disc.management_ip} -> {host}')
                         mods['management_ip'] = host
-                    if site.name != item.site:
-                        changes.append(f'site: {site.name} -> {item.site}')
-                        mods['site_uuid'] = site_names[item.site].uuid
+                    if disc.site_uuid != item._site_uuid:
+                        changes.append(f'site: {disc.site_uuid} -> {item._site_uuid}')
+                        mods['site_uuid'] = item._site_uuid
                     if mods:
                         logging.info('sync_preseem_model(%s): modify %s (%s)', source,
                                      item, ','.join(changes))
@@ -330,13 +434,12 @@ async def sync_preseem_model(preseem_model,
                 else:
                     logging.info('sync_preseem_model(%s): create %s', source, item)
                     if not noact:
-                        site_uuid = site_names[item.site].uuid
                         disc = await call_and_log_error(
                             preseem_model.Element.CreateElementDiscovery(
                                 company_uuid=company_uuid,
                                 name=item.name,
                                 management_ip=item.host,
-                                site_uuid=site_uuid,
+                                site_uuid=item._site_uuid,
                                 source=source,
                                 source_id=item.id,
                                 intent_role_uuid=role.bytes))
@@ -477,12 +580,12 @@ async def sync_preseem_model(preseem_model,
                                 preseem_model.Service.Update,
                                 _service_field_map,
                                 deactivate=deactivate is not False,
+                                cfg=cfg.get('service') if cfg else None,
                                 noact=noact)
 
         for bss_objs, source_type in (aps, AccessPoint), (routers, Router), (olts, Olt):
             if bss_objs is not None:
                 assert isinstance(bss_objs, dict)
-                new_sites = set()
                 for sobj in bss_objs.values():
                     assert isinstance(sobj, source_type)
                     if not sobj.name:  # name is required for ElementDiscovery
@@ -491,32 +594,6 @@ async def sync_preseem_model(preseem_model,
                     if not sobj.host:  # host is required for ElementDiscovery
                         raise ValueError(
                             f'invalid value for {source_type.typename} host')
-                    if sobj.site:
-                        site = site_names.get(sobj.site)
-                        if not site:
-                            new_sites.add(sobj.site)
-                        elif site.inactive:
-                            logging.info('sync_preseem_model(%s): reactivate site "%s"',
-                                         source, sobj.site)
-                            if not noact:
-                                await call_and_log_error(
-                                    preseem_model.Site.Update(company_uuid=company_uuid,
-                                                              uuid=site.uuid,
-                                                              inactive=False))
-                for site_name in new_sites:
-                    logging.info('sync_preseem_model(%s): create site "%s"', source,
-                                 site_name)
-                    if not noact:
-                        site = site_names[site_name] = await call_and_log_error(
-                            preseem_model.Site.Create(company_uuid=company_uuid,
-                                                      name=site_name))
-                        if site:
-                            model_sites[site.uuid] = site_names[site_name] = site
-
-                for sobj in bss_objs.values():
-                    if sobj.site:
-                        site = site_names.get(sobj.site)
-                        sobj._site_uuid = site.uuid
 
                 discs = [
                     x for x in role_discs.get(source_type.role_uuid.bytes) or []
@@ -530,7 +607,7 @@ async def sync_preseem_model(preseem_model,
                     discs,
                     preseem_model.Element.CreateElementDiscovery,
                     preseem_model.Element.UpdateElementDiscovery,
-                    _elem_field_map,
+                    _elem_disc_field_map,
                     deactivate=deactivate is not False,
                     noact=noact)
                 for mobj in discs + new_objs:
@@ -539,3 +616,71 @@ async def sync_preseem_model(preseem_model,
                         sobj._uuid = mobj.uuid
 
         return uuid_map
+
+
+async def sync_preseem_elements(model, company_uuid, source, objs, noact=False):
+    """
+    Update Elements directly in the Preseem model, given source objects such as APs,
+    OLTs, Routers.  This is to support cases when another system such as an NMS is the
+    source of truth for an element's data, and we will not be polling the element
+    directory.  In this case, the element may not even have (or we may not know) a 
+    management IP address, and there is no point in having the poller discover it.
+    """
+    assert isinstance(company_uuid, bytes)
+    assert isinstance(source, str)
+    if source == '':
+        raise ValueError('source cannot be the empty string')
+
+    global sync_lock
+    if sync_lock is None:
+        sync_lock = asyncio.Lock()
+
+    async with sync_lock:
+        # It would be more efficient if we had a way to get elements by source,
+        # but there is no such method, so we get all elements and filter what we need.
+        all_elems, _ = await asyncio.gather(
+            model.Element.List(company_uuid=company_uuid),
+            _sync_sites(model, company_uuid, source, objs.values(), noact=noact))
+
+        # Only keep elements that match a source and poller hash in objs.
+        # This works because we don't currently disable elements anyway, and it
+        # makes sure we're syncing the correct element in the case of a hardware swap.
+        phs = {x.poller_hash: x for x in objs.values()}
+
+        elems = []
+        for elem in all_elems:
+            phelem = phs.get(elem.poller_hash)
+            if elem.source != source or not phelem:
+                continue
+            elems.append(elem)
+
+            # Handle ID changes with the same poller hash.  This needs to be the same
+            # element object.
+            if phelem and phelem.id != elem.source_id:
+                logging.info(
+                    "sync_preseem_element: change element source_id %s: %s -> %s",
+                    UUID(bytes=elem.uuid), elem.source_id, phelem.id)
+                if not noact:
+                    await call_and_log_error(
+                        model.Element.Update(company_uuid=company_uuid,
+                                             uuid=elem.uuid,
+                                             source_id=phelem.id))
+                elem.source_id = phelem.id
+
+        # Sync the objects to the model.  New elements need their status set.
+        init_fields = {'status': ElementStatus.ELEMENT_STATUS_UNKNOWN}
+        new_objs = await _sync_objects(model,
+                                       company_uuid,
+                                       source,
+                                       objs,
+                                       elems,
+                                       model.Element.Create,
+                                       model.Element.Update,
+                                       _elem_field_map,
+                                       deactivate=False,
+                                       init_fields=init_fields,
+                                       noact=noact)
+        for mobj in elems + new_objs:
+            sobj = objs.get(mobj.source_id)
+            if sobj:
+                sobj._uuid = mobj.uuid
