@@ -2,11 +2,15 @@
 Proxy classes for network-poller protobuf messages.
 These classes allow bounds checking and error handling to be centralized.
 """
+import base64
 from collections import namedtuple
+from datetime import timezone
+import json
 import logging
 from preseem_protobuf.network_poller import network_poller_pb2
-# TODO
-# check optional values on Get and return None if not presenet.
+from uuid import UUID
+
+from google.protobuf.json_format import MessageToDict
 
 
 class NetworkElementUpdateException(Exception):
@@ -85,6 +89,36 @@ class BaseWrapperClass:
         attributes_str = ', '.join(f"{attr}={value!r}" for attr, value in attr_values)
         return f"{type(self).__name__}({attributes_str})"
 
+    def pretty(self):
+        """return a pretty-printed string representing this full message."""
+
+        def unencode(d):
+            for a, v in d.items():
+                if a.endswith('uuid'):
+                    try:
+                        d[a] = str(UUID(bytes=base64.b64decode(v)))
+                    except ValueError:
+                        d[a] = v
+                elif a == 'poller_hash':
+                    v = base64.b64decode(v)
+                    if len(v) == 32:
+                        # assume sha256 poller hash
+                        d[a] = v.hex()
+                    else:
+                        # likely an intermediate value, show as a string
+                        # (can't convert bytes to json)
+                        d[a] = v.decode()
+                elif isinstance(v, dict):
+                    unencode(v)
+                elif isinstance(v, list):
+                    for x in v:
+                        if isinstance(x, dict):
+                            unencode(x)
+
+        m = MessageToDict(self.pb, preserving_proto_field_name=True)
+        unencode(m)
+        return json.dumps(m, indent=2)
+
 
 def pb_wrapper(pbtype):
     """Class decorator that sets the PBTYPE and creates an _attrs member."""
@@ -2480,6 +2514,48 @@ class NetworkElement(BaseWrapperClass):
         else:
             self.pb.ClearField('counter_source')
 
+    @property
+    def offline_reason(self):
+        return self.pb.offline_reason if self.pb.HasField('offline_reason') else None
+
+    @offline_reason.setter
+    def offline_reason(self, value):
+        if value is not None:
+            if value > 0:
+                self.pb.offline_reason = value
+            else:  # not a valid enum value
+                self.errctx.add(network_poller_pb2.PollError.ERROR_BAD_DATA_RANGE,
+                                "Invalid value for Networkelement.offline_reason",
+                                value=value,
+                                metric_name='NetworkElement.offline_reason')
+                self.pb.ClearField('offline_reason')
+        else:
+            self.pb.ClearField('offline_reason')
+
+    @property
+    def offline_reason_detail(self):
+        return self.pb.offline_reason_detail if self.pb.HasField(
+            'offline_reason_detail') else None
+
+    @offline_reason_detail.setter
+    def offline_reason_detail(self, value):
+        if value is not None:
+            self.pb.offline_reason_detail = value
+        else:
+            self.pb.ClearField('offline_reason_detail')
+
+    @property
+    def offline_time(self):
+        return self.pb.offline_time.ToDatetime(
+            timezone.utc) if self.pb.HasField('offline_time') else None
+
+    @offline_time.setter
+    def offline_time(self, value):
+        if value is not None:
+            val = self.pb.offline_time.FromDatetime(value)
+        else:
+            self.pb.ClearField('offline_time')
+
 
 @pb_wrapper(network_poller_pb2.NetworkElementUpdate)
 class NetworkElementUpdate(BaseWrapperClass):
