@@ -7,10 +7,11 @@ import unittest
 from uuid import uuid4, UUID
 
 from preseem import FakeNetworkMetadataModel, NetworkMetadataReference, FakePreseemModel
-from preseem.model_sync import sync_preseem_model, ElementSyncItem
+from preseem.model_sync import sync_preseem_elements, sync_preseem_model, ElementSyncItem
 from preseem_protobuf.model.common_pb2 import ElementStatus
 from preseem_protobuf.model.account_pb2 import AccountStatus
 from preseem_protobuf.model.common_pb2 import AccountType
+from preseem_protobuf.model.service_pb2 import SetAttachmentReq
 from preseem.source_model import Attachment, AccessPoint, Olt, Router, Service
 
 Account = namedtuple('Account', ('id', 'type', 'status', 'name', 'first_name', 'last_name'))
@@ -1254,6 +1255,48 @@ class TestModelSync(unittest.TestCase):
             self.assertEqual(msvc.updated_at, model_services[0].updated_at)
         self._await(test())
 
+    def test_service_update_attachment_config(self):
+        """A Service can have its cpe_mac field ignored based on configuration."""
+        async def test():
+            # create a service
+            accounts = {'acct01': Account('acct01', None, AccountStatus.ACCOUNT_STATUS_ACTIVE, 'Account 1', None, None)}
+            packages = {'pkg01': Package('pkg01', 'Package 1', 25000, 5000)}
+            service = Service('service01', account_id='acct01', package_id='pkg01', download_rate=100000, upload_rate=20000, subscriber_identifier='account01', url='https://my.source.url')
+            service.attachment.mac_addresses.append('00:01:02:03:04:05')
+            service.attachment.network_prefixes.append('192.0.2.101')
+            services = {'service01': service}
+            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
+            await self.checkSourceService(service)
+            service.attachment.pb.CopyFrom(service.attachment.pb)
+
+            # change the CPE MAC address externally.
+            svc = (await self.model.Service.List(company_uuid=self.company.uuid))[0]
+            set_att_req = SetAttachmentReq()
+            set_att_req.attachment.CopyFrom(svc.attachment)
+            set_att_req.attachment.cpe_mac_address = "00:01:00:01:00:01"
+            svc = await self.model.Service.Update(company_uuid=self.company.uuid, uuid=svc.uuid, attachment_req=set_att_req)
+
+            # by default, it gets changed back.
+            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services)
+            await self.checkSourceService(service)  # fails since set externally
+
+            # change the CPE MAC address externally.
+            svc = (await self.model.Service.List(company_uuid=self.company.uuid))[0]
+            set_att_req = SetAttachmentReq()
+            set_att_req.attachment.CopyFrom(svc.attachment)
+            set_att_req.attachment.cpe_mac_address = "00:01:00:01:00:01"
+            svc = await self.model.Service.Update(company_uuid=self.company.uuid, uuid=svc.uuid, attachment_req=set_att_req)
+
+            # sync, ignoring this field so a different system is able to change it.
+            # (something else has to change, too, or it won't modify the attachment.)
+            service.attachment.username = "user02"
+            cfg = {'service': {'attachment': {'cpe_mac_address': {'ignore': True}}}}
+
+            await sync_preseem_model(self.model, self.company.uuid, 'test_source', ElementRole.AP, [], accounts=accounts, packages=packages, services=services, cfg=cfg)
+            with self.assertRaises(AssertionError):
+                await self.checkSourceService(service)  # fails since set externally
+        self._await(test())
+
     async def checkSourceElement(self, sobj, role_uuid, deleted=False):
         sites = {}
         for site in await self.model.Site.List(company_uuid=self.company.uuid):
@@ -1455,3 +1498,108 @@ class TestModelSync(unittest.TestCase):
             await self.checkSourceRouter(router, deleted=True)
         self._await(test())
 
+    async def checkSourceElement2(self, sobj):
+        """Check actual elements synced from source objects."""
+        sites = {}
+        for site in await self.model.Site.List(company_uuid=self.company.uuid):
+            if site.name in sites:
+                if isinstance(sites[site.name], list):
+                    sites[site.name].append(site)
+                else:
+                    sites[site.name] = [sites[site.name], site]
+            else:
+                sites[site.name] = site
+        mobjs = {x.source_id: x for x in await self.model.Element.List(company_uuid=self.company.uuid)}
+        mobj = mobjs.get(sobj.id)
+        self.assertIsNotNone(mobj)
+        self.assertEqual(mobj.source_id, sobj.id)
+        self.assertEqual(mobj.name, sobj.name)
+        self.assertEqual(mobj.management_ip, sobj.host)
+        self.assertEqual(mobj.inactive, False)
+        self.assertEqual(mobj.status, ElementStatus.ELEMENT_STATUS_UNKNOWN)
+        self.assertEqual(mobj.poller_hash, sobj.poller_hash)
+        if sobj.site:
+            site = sites.get(sobj.site)
+            if site or not sobj.errors:
+                self.assertIsNotNone(site)
+                self.assertEqual(site.uuid, mobj.site_uuid)
+        self.assertEqual(mobj.uuid, sobj._uuid)
+
+    def test_element_sync(self):
+        """Elements can be directly created and updated."""
+        async def test():
+            # Create two elements.
+            ap_poller_hash = sha256(b"ap1").digest()
+            olt_poller_hash = sha256(b"olt1").digest()
+            ap = AccessPoint('ap01', host='192.0.2.1', name='AP 1', site='Site 1', poller_hash=ap_poller_hash)
+            olt = Olt('olt01', name='OLT 1', site='Site 2', poller_hash=olt_poller_hash)
+            elems = {'ap01': ap, 'olt01': olt}
+            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
+            await self.checkSourceElement2(olt)
+            await self.checkSourceElement2(ap)
+
+            # Update the elements.
+            ap.host = "192.0.2.2"
+            olt.site = None
+            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
+            await self.checkSourceElement2(olt)
+            await self.checkSourceElement2(ap)
+
+            mobjs = await self.model.Element.List(company_uuid=self.company.uuid)
+            self.assertEqual(len(mobjs), 2)
+        self._await(test())
+
+    def test_element_sync_hardware_swap(self):
+        """Hardware swaps are handled by updating the correct poller hash element."""
+        async def test():
+            ap_poller_hash_1 = sha256(b"ap1").digest()
+            ap_poller_hash_2 = sha256(b"ap2").digest()
+            olt_poller_hash = sha256(b"olt1").digest()
+            ap = AccessPoint('ap01', host='192.0.2.1', name='AP 1', site='Site 1', poller_hash=ap_poller_hash_1)
+            olt = Olt('olt01', name='OLT 1', site='Site 2', poller_hash=olt_poller_hash)
+            elems = {'ap01': ap, 'olt01': olt}
+            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
+            await self.checkSourceElement2(olt)
+            await self.checkSourceElement2(ap)
+
+            # The AP gets swapped for new hardware.
+            ap.poller_hash = ap_poller_hash_2
+            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
+            await self.checkSourceElement2(olt)
+            await self.checkSourceElement2(ap)
+
+            mobjs = await self.model.Element.List(company_uuid=self.company.uuid)
+            self.assertEqual(len(mobjs), 3)
+        self._await(test())
+
+    def test_element_sync_source_id_change(self):
+        """If the source ID changes for the same hardware, it should be the same element."""
+        async def test():
+            ap_poller_hash = sha256(b"ap1").digest()
+            olt_poller_hash = sha256(b"olt1").digest()
+            ap = AccessPoint('ap01', host='192.0.2.1', name='AP 1', site='Site 1', poller_hash=ap_poller_hash)
+            olt = Olt('olt01', name='OLT 1', site='Site 2', poller_hash=olt_poller_hash)
+            elems = {'ap01': ap, 'olt01': olt}
+            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
+            await self.checkSourceElement2(olt)
+            await self.checkSourceElement2(ap)
+
+            # The AP ID get changed
+            ap = AccessPoint('ap02', host='192.0.2.1', name='AP 1', site='Site 1', poller_hash=ap_poller_hash)
+            elems = {'ap02': ap, 'olt01': olt}
+            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
+            await self.checkSourceElement2(olt)
+            await self.checkSourceElement2(ap)
+            mobjs = await self.model.Element.List(company_uuid=self.company.uuid)
+            self.assertEqual(len(mobjs), 2)
+
+            # Change something else
+            ap.name = "AP 2"
+            elems = {'ap02': ap, 'olt01': olt}
+            await sync_preseem_elements(self.model, self.company.uuid, 'test_source', elems)
+            await self.checkSourceElement2(olt)
+            await self.checkSourceElement2(ap)
+
+            mobjs = await self.model.Element.List(company_uuid=self.company.uuid)
+            self.assertEqual(len(mobjs), 2)
+        self._await(test())
