@@ -1,4 +1,5 @@
 """Test NetworkElementUpdate framework"""
+from datetime import datetime, timezone
 import unittest
 from hashlib import sha256
 from uuid import uuid4
@@ -393,6 +394,31 @@ class TestNetworkElementUpdate(unittest.TestCase):
         ne.counter_source = "snmp"
         self.assertEqual(ne.counter_source, "snmp")
 
+        self.assertEqual(ne.offline_reason, None)
+        ne.offline_reason = network_poller_pb2.OfflineReason.OFFLINE_REASON_POWER_OFF
+        self.assertEqual(ne.offline_reason, network_poller_pb2.OfflineReason.OFFLINE_REASON_POWER_OFF)
+        ne.offline_reason = None
+        self.assertEqual(ne.offline_reason, None)
+        ne.offline_reason = network_poller_pb2.OfflineReason.OFFLINE_REASON_OTHER
+        self.assertEqual(ne.offline_reason, network_poller_pb2.OfflineReason.OFFLINE_REASON_OTHER)
+
+        self.assertEqual(ne.offline_reason_detail, None)
+        ne.offline_reason_detail = "dying gasp"
+        self.assertEqual(ne.offline_reason_detail, "dying gasp")
+        ne.offline_reason_detail = None
+        self.assertEqual(ne.offline_reason_detail, None)
+        ne.offline_reason_detail = "unknown"
+        self.assertEqual(ne.offline_reason_detail, "unknown")
+
+        ut_time = datetime(2025, 7, 14, 12, 34, 56, tzinfo=timezone.utc)
+        self.assertEqual(ne.offline_time, None)
+        ne.offline_time = ut_time
+        self.assertEqual(ne.offline_time, ut_time)
+        ne.offline_time = None
+        self.assertEqual(ne.offline_time, None)
+        ne.offline_time = ut_time
+        self.assertEqual(ne.offline_time, ut_time)
+
         # Interface
         intf = ne.interfaces.add()
         intf.id = '1'
