@@ -753,3 +753,15 @@ class TestSourceModel(unittest.TestCase):
         site.network_prefixes.append('192.0.2.1')
         self.assertEqual(site.network_prefixes[0], '192.0.2.1')
 
+    def test_element_poller_hash(self):
+        """Test field checking of element.poller_hash"""
+        elem = Olt("elem1", poller_hash=None)
+        self.assertEqual(elem.poller_hash, None)
+        elem = Olt("elem1", poller_hash="")
+        self.assertEqual(elem.poller_hash, None)
+        self.assertError(elem, 'poller_hash', "")
+        elem.poller_hash = b'asdf'
+        self.assertError(elem, 'poller_hash', b"asdf")
+        poller_hash = sha256(b"asdf").digest()
+        elem.poller_hash = poller_hash
+        self.assertEqual(elem.poller_hash, poller_hash)
